#pragma semicolon 1
#define CONVAR_PREFIX "sm_sessionflags"
#define DEFAULT_UPDATE_SETTING "2"
#define UPD_LIBFUNC
#include <ddhoward_updater>
#pragma newdecls required;

#define BAD_COMMAND_SYNTAX "[SM] Usage: sm_flags <#userid|name> [+/-/%%/=/_] <flags>"
#define ALL_FLAGS_ENABLED RoundToFloor(Pow(float(2), float(AdminFlags_TOTAL)) - 1.0)
#define ALL_FLAGS_STRING "abcdefghijklmnopqrstz" //will not cause issues if the server has modified what symbols represent the flags

int realFlags[MAXPLAYERS + 1];
bool flagsChanged[MAXPLAYERS + 1];


ConVar hcvar_logging;

public Plugin myinfo = {
	name = "Session Flags",
	author = "Derek D. Howard (ddhoward)",
	description = "Set users flags during the current session. Loosely based off of \"Niggy's Session Flags\"",
	version = "18.0114.0",
	url = "https://forums.alliedmods.net/showthread.php?t=247263"
};

public void OnPluginStart() {
	LoadTranslations("common.phrases");
	RegAdminCmd("sm_flags", commandFlags, ADMFLAG_ROOT, "Changes clients admin flags on the fly.");
	hcvar_logging = CreateConVar("sm_sessionflags_logging", "2", "(0/1/2) 0 = no logging, 1 = basic logging, 2 = verbose logging");
}

public void OnClientDisconnect(int client) {
	realFlags[client] = 0;
	flagsChanged[client] = false;
}

public Action commandFlags(int client, int args) {
	
	//make sure that they are using the right number of args
	if (args > 3) { //too many arguments
		ReplyToCommand(client, BAD_COMMAND_SYNTAX);
		return Plugin_Handled;
	}
	if (args < 2) {
		//too few arguments for flag alteration, user wants to see flags rather than change them
		
		if (!CheckCommandAccess(client, "sm_who", ADMFLAG_GENERIC) || (args == 0 && client == 0)) {
			//client either doesn't have access to sm_who, or is the console trying to view non-existent flags on "self"
			ReplyToCommand(client, BAD_COMMAND_SYNTAX);
			return Plugin_Handled;
		}
		
		int target = -1;
		if (args == 0) target = client; //no arguments = target is self
		else {
			char target_string[65];
			GetCmdArg(1, target_string, sizeof(target_string));
			target = FindTarget(client, target_string, true, false);
		}
		if (target == -1) {
			ReplyToCommand(client, BAD_COMMAND_SYNTAX);
			return Plugin_Handled;
		}
		
		char targetString[MAX_NAME_LENGTH+3];
		if (client == target) { targetString = "Your"; }
		else {
			Format(targetString, sizeof(targetString), "%N's", target);
		}
		
		char currentFlags[AdminFlags_TOTAL + 1];
		FlagbitsToFlagstring(GetUserFlagBits(target), currentFlags);
		if (flagsChanged[target]) {
			char originalFlags[AdminFlags_TOTAL + 1];
			FlagbitsToFlagstring(realFlags[target], originalFlags);
			ReplyToCommand(client, "[SM] %s current flags are %s and original flags are %s.", targetString, currentFlags, originalFlags);
		}
		else {
			ReplyToCommand(client, "[SM] %s current flags are %s.", targetString, currentFlags);
		}
		return Plugin_Handled;
	}

	
	//store whether the user wants to add, remove, toggle, reset, etc.
	char mode_string[2];
	int mode;

	char flagstring[AdminFlags_TOTAL+1];
	char action[14];
	char action2[5];

	
	if (args == 2) { 
		//no mode specified, second argument is flags
		GetCmdArg(2, flagstring, sizeof(flagstring));
		action = "gave flags";
		action2 = "to";
	} 
	else if (args == 3) { //a mode and flags were both specified
		GetCmdArg(3, flagstring, sizeof(flagstring));
		GetCmdArg(2, mode_string, sizeof(mode_string));
		switch (mode_string[0]) {
			case '+': {
				mode = 0;
				action = "gave flags";
				action2 = "to";
			}
			case '-': {
				mode = 1;
				action = "removed flags";
				action2 = "from";
			}
			case '%': {
				mode = 2;
				action = "toggled flags";
				action2 = "on";
			}
			case '=': {
				mode = 3;
				action = "set flags to";
				action2 = "on";
			}
			case '_': {
				mode = 4;
				action = "reset flags";
				action2 = "on";
			}
			default: {
				ReplyToCommand(client, BAD_COMMAND_SYNTAX);
				return Plugin_Handled;
			}
		}
	}
	
	//get targets, duh
	char target_string[65];
	GetCmdArg(1, target_string, sizeof(target_string));
	char target_name[MAX_TARGET_LENGTH];
	int target_list[MAXPLAYERS];
	int target_count;
	bool tn_is_ml;
	if ((target_count = ProcessTargetString(
			target_string,
			client,
			target_list,
			MAXPLAYERS,
			0,
			target_name,
			sizeof(target_name),
			tn_is_ml)) <= 0) {
		ReplyToTargetError(client, target_count);
		return Plugin_Handled;
	}

	//get the command user's flags
	int adminFlags;
	if (CheckCommandAccess(client, "", ADMFLAG_ROOT, true)
	|| (mode == 4 && CheckCommandAccess(client, "sm_sessionflags_reset", ADMFLAG_ROOT, true))) {
		adminFlags = ALL_FLAGS_ENABLED; }
	else { adminFlags = GetUserFlagBits(client); }


	//clean up the flagstring input and make it flagbits
	int requestedFlags;
	if (flagstring[0] == '@') {
		requestedFlags = ALL_FLAGS_ENABLED;
	}
	else {
		AdminFlag useless;
		for (int i = 0; i < strlen(flagstring); i++) {
			flagstring[i] = CharToLower(flagstring[i]);
			//upper case flags don't work
			if (!FindFlagByChar(flagstring[i], useless)) {
				//get rid of characters that aren't valid flags
				char temp[1];
				temp[0] = flagstring[i];
				ReplaceStringEx(flagstring, sizeof(flagstring), temp, "", 1, 0);
				i--;
				continue;
			}
		}
		requestedFlags = ReadFlagString(flagstring);
	}
	
	//stop everything if the admin hasn't specified any flags that he can actually affect
	if (requestedFlags & adminFlags == 0 && mode != 3) {
		ReplyToCommand(client, "[SM] No valid flags were specified!");
		return Plugin_Handled;
	}
	//when logging or showing activity, don't show flags that the admin can't affect
	FlagbitsToFlagstring(requestedFlags & adminFlags, flagstring);
	if (StrEqual(flagstring, ALL_FLAGS_STRING)) { flagstring = "all flags"; }

	//create the bool arrays
	bool oldFlags_array[AdminFlags_TOTAL];
	bool requestedFlags_array[AdminFlags_TOTAL];
	bool adminFlags_array[AdminFlags_TOTAL];
	bool newFlags_array[AdminFlags_TOTAL];
	
	//populate the arrays for the command user
	FlagBitsToBitArray(requestedFlags, requestedFlags_array, AdminFlags_TOTAL);
	FlagBitsToBitArray(adminFlags, adminFlags_array, AdminFlags_TOTAL);
	
	char oldString[AdminFlags_TOTAL + 1];
	char newString[AdminFlags_TOTAL + 1];
	
	//is set to true if a player's flags is actually changed
	//if it stays false, no information is printed or logged
	bool significant;

	//loop through all targets
	for (int i = 0; i < target_count; i++) {
		int target = target_list[i];
		int oldFlags = GetUserFlagBits(target);

		if (!flagsChanged[target]) {
			//preserve the player's "real" flags for use in resetting later
			//this is done now rather than when the client joins in case other
			//plugins have been making changes to his/her flags
			realFlags[target] = oldFlags;
			flagsChanged[target] = true;
		}
		
		FlagBitsToBitArray(oldFlags, oldFlags_array, AdminFlags_TOTAL);
		FlagBitsToBitArray(oldFlags, newFlags_array, AdminFlags_TOTAL);
		int newFlags;
		
		switch (mode) {
			case 0: { //add flags
				for (int f = 0; f < AdminFlags_TOTAL; f++) {
					if (adminFlags_array[f] && requestedFlags_array[f]) { newFlags_array[f] = true; }
				}
			}
			case 1: { //remove flags
				for (int f = 0; f < AdminFlags_TOTAL; f++) {
					if (adminFlags_array[f] && requestedFlags_array[f]) { newFlags_array[f] = false; }
				}
			}
			case 2: { //toggle flags
				for (int f = 0; f < AdminFlags_TOTAL; f++) {
					if (adminFlags_array[f] && requestedFlags_array[f]) { newFlags_array[f] = !newFlags_array[f]; }
				}
			}
			case 3: { //set exact flags
				for (int f = 0; f < AdminFlags_TOTAL; f++) {
					if (adminFlags_array[f]) { newFlags_array[f] = requestedFlags_array[f]; }
				}
			}
			case 4: { //reset flags to normal

				//put the target's original flags into a boolean array
				bool realFlags_array[AdminFlags_TOTAL];
				FlagBitsToBitArray(realFlags[target], realFlags_array, AdminFlags_TOTAL);

				for (int f = 0; f < AdminFlags_TOTAL; f++) {
					if (adminFlags_array[f] && requestedFlags_array[f]) { newFlags_array[f] = realFlags_array[f]; }
				}				
			}
		}
		SetUserFlagBits(target, (newFlags = FlagBitArrayToBits(newFlags_array, AdminFlags_TOTAL))); //do it
		
		//if the target's flags are the same as realFlags, unmark flags as changed
		if (newFlags == realFlags[target]) { flagsChanged[target] = false; }
		
		if (newFlags != oldFlags) {
			significant = true;
			if ((target_count == 1 && hcvar_logging.IntValue == 1) || hcvar_logging.IntValue > 1) {
				FlagbitsToFlagstring(newFlags, newString);
				FlagbitsToFlagstring(oldFlags, oldString);
				LogAction(client, target, "\"%L\" %s \"%s\" %s \"%L\". (%s to %s)", client, action, flagstring, action2, target, oldString, newString);
			}
		}
	}
	
	if (significant) {
		ShowActivity2(client, "[SM] ", "%s \"%s\" %s %s", action, flagstring, action2, target_name);
		if (target_count > 1 && hcvar_logging.IntValue == 1) {
			LogAction(client, -1, "\"%L\" %s \"%s\" %s %s.", client, action, flagstring, action2, target_name);
		}
	}
	else { ReplyToCommand(client, "[SM] No players were affected."); }	
	
	return Plugin_Handled;
}

stock void FlagbitsToFlagstring(int flagbits, char flagstring[AdminFlags_TOTAL + 1]) {
	if (flagbits == 0) {
		flagstring = "0";
		return;
	}
	
	flagstring = "";
	int flagstringPosition;
	
	bool flagBitArray[AdminFlags_TOTAL];
	FlagBitsToBitArray(flagbits, flagBitArray, sizeof(flagBitArray));
	bool addZ;
	for (int i = 0; i < sizeof(flagBitArray); i++) {
		flagbits = 0;
		if (flagBitArray[i]) {
			flagbits = (1 << i);
			AdminFlag flag;
			if (!BitToFlag(flagbits, flag)) { continue; }
			int character;
			if (!FindFlagChar(flag, character)) { continue; }
			if (character == 'z') { addZ = true; }
			else {
				flagstring[flagstringPosition] = character;
				flagstringPosition++;
			}
		}
	}
	if (addZ) {
		flagstring[flagstringPosition] = 'z';
		flagstringPosition++;
	}
	flagstring[flagstringPosition] = '\0';
}